@@ -26,6 +26,9 @@
 #include <string.h>
 #include <getopt.h>
 #include <errno.h>
+#include <ctype.h>
+#include <regex.h>
+#include <inttypes.h>
 #include <sys/fcntl.h>
 #include <sys/stat.h>
 
@@ -37,27 +40,28 @@
 #include <osmocom/core/application.h>
 #include <osmocom/core/select.h>
 #include <osmocom/core/timer.h>
-#include <openbsc/ipaccess.h>
-#include <openbsc/common_bsc.h>
+#include <osmocom/bsc/ipaccess.h>
 #include <osmocom/abis/e1_input.h>
-#include <openbsc/abis_nm.h>
-#include <openbsc/signal.h>
-#include <openbsc/debug.h>
-#include <openbsc/network_listen.h>
+#include <osmocom/bsc/abis_nm.h>
+#include <osmocom/bsc/signal.h>
+#include <osmocom/bsc/debug.h>
+#include <osmocom/bsc/network_listen.h>
 #include <osmocom/abis/ipaccess.h>
-#include <openbsc/gsm_data.h>
-#include <openbsc/abis_nm.h>
-#include <openbsc/signal.h>
-#include <openbsc/debug.h>
-#include <openbsc/network_listen.h>
+#include <osmocom/bsc/gsm_data.h>
+#include <osmocom/bsc/abis_nm.h>
+#include <osmocom/bsc/signal.h>
+#include <osmocom/bsc/debug.h>
+#include <osmocom/bsc/network_listen.h>
 #include <osmocom/core/talloc.h>
 #include <osmocom/abis/abis.h>
 #include <osmocom/gsm/protocol/gsm_12_21.h>
+#include <osmocom/bsc/bss.h>
 
 struct gsm_network *bsc_gsmnet;
 
 static int net_listen_testnr;
 static int restart;
+static bool get_attr;
 static char *prim_oml_ip;
 static char *bts_ip_addr, *bts_ip_mask, *bts_ip_gw;
 static char *unit_id;
@@ -70,16 +74,12 @@ static int dump_files = 0;
 static char *firmware_analysis = NULL;
 static int found_trx = 0;
 static int loop_tests = 0;
+static bool quiet = false;
 
 static void *tall_ctx_config = NULL;
 static struct abis_nm_sw_desc *sw_load1 = NULL;
 static struct abis_nm_sw_desc *sw_load2 = NULL;
 
-/*
-static uint8_t prim_oml_attr[] = { 0x95, 0x00, 7, 0x88, 192, 168, 100, 11, 0x00, 0x00 };
-static uint8_t unit_id_attr[] = { 0x91, 0x00, 9, '2', '3', '4', '2', '/' , '0', '/', '0', 0x00 };
-*/
-
 extern int ipaccess_fd_cb(struct osmo_fd *bfd, unsigned int what);
 extern struct e1inp_line_ops ipaccess_e1inp_line_ops;
 
@@ -117,6 +117,7 @@ static int ipaccess_connect(struct e1inp_line *line, struct sockaddr_in *sa)
 
 	ret = osmo_fd_register(bfd);
 	if (ret < 0) {
+		LOGP(DLINP, LOGL_ERROR, "unable to register socket fd\n");
 		close(bfd->fd);
 		return ret;
 	}
@@ -191,7 +192,7 @@ static void check_restart_or_exit(struct gsm_bts_trx *trx)
 static int ipacc_msg_ack(uint8_t mt, struct gsm_bts_trx *trx)
 {
 	if (sw_load_state == 1) {
-		fprintf(stderr, "The new software is activaed.\n");
+		fprintf(stderr, "The new software is activated.\n");
 		check_restart_or_exit(trx);
 	} else if (oml_state == 1) {
 		fprintf(stderr, "Set the NV Attributes.\n");
@@ -237,7 +238,7 @@ static int nwl_sig_cb(unsigned int subsys, unsigned int signal,
 			/* Create whitelist from results */
 			physconf_len = build_physconf(physconf_buf,
 						      &trx->ipaccess.rxlev_stat);
-			/* Start next test abbout BCCH channel usage */
+			/* Start next test about BCCH channel usage */
 			ipac_nwl_test_start(trx, NM_IPACC_TESTNO_BCCH_CHAN_USAGE,
 					    physconf_buf, physconf_len);
 			break;
@@ -267,6 +268,33 @@ static int nwl_sig_cb(unsigned int subsys, unsigned int signal,
 	return 0;
 }
 
+static int print_attr_rep(struct msgb *mb)
+{
+	/* Parse using nanoBTS own formatting for Get Attribute Response */
+	struct abis_om_hdr *oh = msgb_l2(mb);
+	struct abis_om_fom_hdr *foh = msgb_l3(mb);
+	struct e1inp_sign_link *sign_link = mb->dst;
+	struct gsm_bts_trx *trx = sign_link->trx;
+	struct gsm_bts *bts = trx->bts;
+	struct tlv_parsed tp;
+	struct in_addr ia = {0};
+	char oml_ip[20] = {0};
+	uint16_t oml_port = 0;
+	char unit_id[40]  = {0};
+
+
+	abis_nm_tlv_parse(&tp, bts, foh->data, oh->length-sizeof(*foh));
+
+	abis_nm_tlv_attr_primary_oml(&tp, &ia, &oml_port);
+	osmo_strlcpy(oml_ip, inet_ntoa(ia), sizeof(oml_ip));
+
+	abis_nm_tlv_attr_unit_id(&tp, unit_id,  sizeof(unit_id));
+
+	fprintf(stdout, "{ \"primary_oml_ip\": \"%s\", \"primary_oml_port\": %" PRIu16 ", \"unit_id\": \"%s\" }\n",
+		oml_ip, oml_port, unit_id);
+	return 0;
+}
+
 static int nm_state_event(int evt, uint8_t obj_class, void *obj,
 			  struct gsm_nm_state *old_state, struct gsm_nm_state *new_state,
 			  struct abis_om_obj_inst *obj_inst);
@@ -276,6 +304,7 @@ static int nm_sig_cb(unsigned int subsys, unsigned int signal,
 {
 	struct ipacc_ack_signal_data *ipacc_data;
 	struct nm_statechg_signal_data *nsd;
+	struct msgb *oml_msg;
 
 	switch (signal) {
 	case S_NM_IPACC_NACK:
@@ -285,11 +314,13 @@ static int nm_sig_cb(unsigned int subsys, unsigned int signal,
 		ipacc_data = signal_data;
 		return ipacc_msg_ack(ipacc_data->msg_type, ipacc_data->trx);
 	case S_NM_IPACC_RESTART_ACK:
-		printf("The BTS has acked the restart. Exiting.\n");
+		if (!quiet)
+			printf("The BTS has acked the restart. Exiting.\n");
 		exit(0);
 		break;
 	case S_NM_IPACC_RESTART_NACK:
-		printf("The BTS has nacked the restart. Exiting.\n");
+		if (!quiet)
+			printf("The BTS has nacked the restart. Exiting.\n");
 		exit(0);
 		break;
 	case S_NM_STATECHG_OPER:
@@ -298,6 +329,11 @@ static int nm_sig_cb(unsigned int subsys, unsigned int signal,
 		nm_state_event(signal, nsd->obj_class, nsd->obj, nsd->old_state,
 				nsd->new_state, nsd->obj_inst);
 		break;
+	case S_NM_GET_ATTR_REP:
+		fprintf(stderr, "Received SIGNAL S_NM_GET_ATTR_REP\n");
+		oml_msg = signal_data;
+		print_attr_rep(oml_msg);
+		exit(0);
 	default:
 		break;
 	}
@@ -321,7 +357,8 @@ static int swload_cbfn(unsigned int hook, unsigned int event, struct msgb *_msg,
 
 	switch (event) {
 	case NM_MT_LOAD_INIT_ACK:
-		fprintf(stdout, "Software Load Initiate ACK\n");
+		if (!quiet)
+			fprintf(stdout, "Software Load Initiate ACK\n");
 		break;
 	case NM_MT_LOAD_INIT_NACK:
 		fprintf(stderr, "ERROR: Software Load Initiate NACK\n");
@@ -347,7 +384,9 @@ static int swload_cbfn(unsigned int hook, unsigned int event, struct msgb *_msg,
 		msg->l2h[0] = NM_ATT_IPACC_CUR_SW_CFG;
 		msg->l2h[1] = msgb_l3len(msg) >> 8;
 		msg->l2h[2] = msgb_l3len(msg) & 0xff;
-		printf("Foo l2h: %p l3h: %p... length l2: %u  l3: %u\n", msg->l2h, msg->l3h, msgb_l2len(msg), msgb_l3len(msg));
+		if (!quiet)
+			printf("Foo l2h: %p l3h: %p... length l2: %u  l3: %u\n",
+			       msg->l2h, msg->l3h, msgb_l2len(msg), msgb_l3len(msg));
 		abis_nm_ipaccess_set_nvattr(trx, msg->l2h, msgb_l2len(msg));
 		msgb_free(msg);
 		break;
@@ -363,7 +402,7 @@ static int swload_cbfn(unsigned int hook, unsigned int event, struct msgb *_msg,
 		break;
 	case NM_MT_LOAD_SEG_ACK:
 		percent = abis_nm_software_load_status(trx->bts);
-		if (percent > percent_old)
+		if (!quiet && percent > percent_old)
 			printf("Software Download Progress: %d%%\n", percent);
 		percent_old = percent;
 		break;
@@ -407,7 +446,7 @@ static void nv_put_prim_oml(struct msgb *nmsg, uint32_t ip, uint16_t port)
 	msgb_put_u8(nmsg, NM_ATT_IPACC_PRIM_OML_CFG_LIST);
 	msgb_put_u16(nmsg, len);
 
-	msgb_put_u8(nmsg, 0x88);
+	msgb_put_u8(nmsg, NM_ATT_IPACC_PRIM_OML_CFG);
 
 	/* IP address */
 	msgb_put_u32(nmsg, ip);
@@ -491,18 +530,25 @@ static int ipa_nvflag_set(uint16_t *flags, uint16_t *mask, const char *name, int
 
 static void bootstrap_om(struct gsm_bts_trx *trx)
 {
-	struct msgb *nmsg = msgb_alloc(1024, "nested msgb");
+	struct msgb *nmsg_get = msgb_alloc(1024, "nested get msgb");
+	struct msgb *nmsg_set = msgb_alloc(1024, "nested set msgb");
 	int need_to_set_attr = 0;
 	int len;
 
-	printf("OML link established using TRX %d\n", trx->nr);
+	if (!quiet)
+		printf("OML link established using TRX %d\n", trx->nr);
 
+	if (get_attr) {
+		msgb_put_u8(nmsg_get, NM_ATT_IPACC_PRIM_OML_CFG);
+		msgb_put_u8(nmsg_get, NM_ATT_IPACC_UNIT_ID);
+	}
 	if (unit_id) {
 		len = strlen(unit_id);
-		if (len > nmsg->data_len-10)
+		if (len > nmsg_set->data_len-10)
 			goto out_err;
-		printf("setting Unit ID to '%s'\n", unit_id);
-		nv_put_unit_id(nmsg, unit_id);
+		if (!quiet)
+			printf("setting Unit ID to '%s'\n", unit_id);
+		nv_put_unit_id(nmsg_set, unit_id);
 		need_to_set_attr = 1;
 	}
 	if (prim_oml_ip) {
@@ -514,14 +560,16 @@ static void bootstrap_om(struct gsm_bts_trx *trx)
 			goto out_err;
 		}
 
-		printf("setting primary OML link IP to '%s'\n", inet_ntoa(ia));
-		nv_put_prim_oml(nmsg, ntohl(ia.s_addr), 0);
+		if (!quiet)
+			printf("setting primary OML link IP to '%s'\n", inet_ntoa(ia));
+		nv_put_prim_oml(nmsg_set, ntohl(ia.s_addr), 0);
 		need_to_set_attr = 1;
 	}
 	if (nv_mask) {
-		printf("setting NV Flags/Mask to 0x%04x/0x%04x\n",
-			nv_flags, nv_mask);
-		nv_put_flags(nmsg, nv_flags, nv_mask);
+		if (!quiet)
+			printf("setting NV Flags/Mask to 0x%04x/0x%04x\n",
+			       nv_flags, nv_mask);
+		nv_put_flags(nmsg_set, nv_flags, nv_mask);
 		need_to_set_attr = 1;
 	}
 	if (bts_ip_addr && bts_ip_mask) {
@@ -539,8 +587,9 @@ static void bootstrap_om(struct gsm_bts_trx *trx)
 			goto out_err;
 		}
 
-		printf("setting static IP Address/Mask\n");
-		nv_put_ip_if_cfg(nmsg, ntohl(ia_addr.s_addr), ntohl(ia_mask.s_addr));
+		if (!quiet)
+			printf("setting static IP Address/Mask\n");
+		nv_put_ip_if_cfg(nmsg_set, ntohl(ia_addr.s_addr), ntohl(ia_mask.s_addr));
 		need_to_set_attr = 1;
 	}
 	if (bts_ip_gw) {
@@ -552,24 +601,33 @@ static void bootstrap_om(struct gsm_bts_trx *trx)
 			goto out_err;
 		}
 
-		printf("setting static IP Gateway\n");
+		if (!quiet)
+			printf("setting static IP Gateway\n");
 		/* we only set the default gateway with zero addr/mask */
-		nv_put_gw_cfg(nmsg, 0, 0, ntohl(ia_gw.s_addr));
+		nv_put_gw_cfg(nmsg_set, 0, 0, ntohl(ia_gw.s_addr));
 		need_to_set_attr = 1;
 	}
 
+	if (get_attr) {
+		fprintf(stderr, "getting Attributes (%d): %s\n", nmsg_get->len, osmo_hexdump(msgb_data(nmsg_get), msgb_length(nmsg_get)));
+		abis_nm_get_attr(trx->bts, NM_OC_BASEB_TRANSC, 0, trx->nr, 0xff, nmsg_get->head, nmsg_get->len);
+		oml_state = 1;
+	}
+
 	if (need_to_set_attr) {
-		abis_nm_ipaccess_set_nvattr(trx, nmsg->head, nmsg->len);
+		abis_nm_ipaccess_set_nvattr(trx, nmsg_set->head, nmsg_set->len);
 		oml_state = 1;
 	}
 
 	if (restart && !prim_oml_ip && !software) {
-		printf("restarting BTS\n");
+		if (!quiet)
+			printf("restarting BTS\n");
 		abis_nm_ipaccess_restart(trx);
 	}
 
 out_err:
-	msgb_free(nmsg);
+	msgb_free(nmsg_get);
+	msgb_free(nmsg_set);
 }
 
 static int nm_state_event(int evt, uint8_t obj_class, void *obj,
@@ -592,7 +650,8 @@ static int nm_state_event(int evt, uint8_t obj_class, void *obj,
 					    phys_conf_min, sizeof(phys_conf_min));
 		else if (software) {
 			int rc;
-			printf("Attempting software upload with '%s'\n", software);
+			if (!quiet)
+				printf("Attempting software upload with '%s'\n", software);
 			rc = abis_nm_software_load(trx->bts, trx->nr, software, 19, 0, swload_cbfn, trx);
 			if (rc < 0) {
 				fprintf(stderr, "Failed to start software load\n");
@@ -632,7 +691,7 @@ static int find_sw_load_params(const char *filename)
 	INIT_LLIST_HEAD(entry);
 
 	fd = open(filename, O_RDONLY);
-	if (!fd) {
+	if (fd < 0) {
 		perror("nada");
 		return -1;
 	}
@@ -725,9 +784,10 @@ static void analyze_firmware(const char *filename)
 	entry = talloc_zero(tall_firm_ctx, struct llist_head);
 	INIT_LLIST_HEAD(entry);
 
-	printf("Opening possible firmware '%s'\n", filename);
+	if (!quiet)
+		printf("Opening possible firmware '%s'\n", filename);
 	fd = open(filename, O_RDONLY);
-	if (!fd) {
+	if (fd < 0) {
 		perror("nada");
 		return;
 	}
@@ -781,6 +841,28 @@ static void analyze_firmware(const char *filename)
 	talloc_free(tall_firm_ctx);
 }
 
+static bool check_unitid_fmt(const char* unit_id)
+{
+	regex_t regexp;
+	int rc;
+
+	if (strlen(unit_id) < 5)
+		goto wrong_fmt;
+
+	rc = regcomp(&regexp, "^[0-9]+/[0-9]+/[0-9]+$", REG_EXTENDED | REG_NOSUB);
+	OSMO_ASSERT(!rc);
+
+	rc = regexec(&regexp, unit_id, 0, NULL, 0);
+	regfree(&regexp);
+
+	if (rc == 0)
+		return true;
+
+wrong_fmt:
+	fprintf(stderr, "ERROR: unit-id wrong format. Must be '\\d+/\\d+/\\d+'\n");
+	return false;
+}
+
 static void print_usage(void)
 {
 	printf("Usage: ipaccess-config IP_OF_BTS\n");
@@ -796,11 +878,12 @@ static void print_help(void)
 	printf("Commands for writing to the BTS:\n");
 	printf("  -u --unit-id UNIT_ID\t\tSet the Unit ID of the BTS\n");
 	printf("  -o --oml-ip IP\t\tSet primary OML IP (IP of your BSC)\n");
+	printf("  -G --get-attr\t\t\tGet several attributes from BTS\n");
 	printf("  -i --ip-address IP/MASK\tSet static IP address + netmask of BTS\n");
 	printf("  -g --ip-gateway IP\t\tSet static IP gateway of BTS\n");
 	printf("  -r --restart\t\t\tRestart the BTS (after other operations)\n");
 	printf("  -n --nvram-flags FLAGS/MASK\tSet NVRAM attributes\n");
-	printf("  -S --nvattr-set FLAG\tSet one additional NVRAM attribute\n");
+	printf("  -S --nvattr-set FLAG\t\tSet one additional NVRAM attribute\n");
 	printf("  -U --nvattr-unset FLAG\tSet one additional NVRAM attribute\n");
 	printf("  -l --listen TESTNR\t\tPerform specified test number\n");
 	printf("  -L --Listen TEST_NAME\t\tPerform specified test\n");
@@ -813,6 +896,7 @@ static void print_help(void)
 	printf("  -f --firmware FIRMWARE\tProvide firmware information\n");
 	printf("  -w --write-firmware\t\tThis will dump the firmware parts to the filesystem. Use with -f.\n");
 	printf("  -p --loop\t\t\tLoop the tests executed with the --listen command.\n");
+	printf("  -q --quiet\t\t\tAvoid printing unformatted logging to stdout, useful with -G.\n");
 }
 
 static void print_value_string(const struct value_string *val, int size)
@@ -836,29 +920,41 @@ static void print_options(void)
 	print_value_string(&ipa_test_strs[0], ARRAY_SIZE(ipa_test_strs));
 }
 
-extern void bts_model_nanobts_init();
+static const struct log_info_cat log_categories[] = {
+	[DNM] = {
+		.name = "DNM",
+		.description = "A-bis Network Management / O&M (NM/OML)",
+		.color = "\033[1;36m",
+		.loglevel = LOGL_DEBUG,
+		.enabled = 1,
+	},
+};
+
+static const struct log_info log_info = {
+	.cat = log_categories,
+	.num_cat = ARRAY_SIZE(log_categories),
+};
 
 int main(int argc, char **argv)
 {
 	struct gsm_bts *bts;
 	struct sockaddr_in sin;
+	char *bts_ip;
 	int rc, option_index = 0, stream_id = 0xff;
 
 	tall_ctx_config = talloc_named_const(NULL, 0, "ipaccess-config");
+	tall_bsc_ctx = tall_ctx_config;
 	msgb_talloc_ctx_init(tall_ctx_config, 0);
 
-	osmo_init_logging(&log_info);
-	log_parse_category_mask(osmo_stderr_target, "DNM,0");
+	osmo_init_logging2(tall_ctx_config, &log_info);
 	bts_model_nanobts_init();
 
-	printf("ipaccess-config (C) 2009-2010 by Harald Welte and others\n");
-	printf("This is FREE SOFTWARE with ABSOLUTELY NO WARRANTY\n\n");
-
 	while (1) {
 		int c;
 		unsigned long ul;
 		char *slash;
 		static struct option long_options[] = {
+			{ "get-attr", 0, 0, 'G' },
 			{ "unit-id", 1, 0, 'u' },
 			{ "oml-ip", 1, 0, 'o' },
 			{ "ip-address", 1, 0, 'i' },
@@ -877,17 +973,23 @@ int main(int argc, char **argv)
 			{ "write-firmware", 0, 0, 'w' },
 			{ "disable-color", 0, 0, 'c'},
 			{ "loop", 0, 0, 'p' },
+			{ "quiet", 0, 0, 'q' },
 			{ 0, 0, 0, 0 },
 		};
 
-		c = getopt_long(argc, argv, "u:o:i:g:rn:S:U:l:L:hs:d:f:wcpH", long_options,
+		c = getopt_long(argc, argv, "Gu:o:i:g:rn:S:U:l:L:hs:d:f:wcpqH", long_options,
 				&option_index);
 
 		if (c == -1)
 			break;
 
 		switch (c) {
+		case 'G':
+			get_attr = true;
+			break;
 		case 'u':
+			if (!check_unitid_fmt(optarg))
+				exit(2);
 			unit_id = optarg;
 			break;
 		case 'o':
@@ -957,6 +1059,9 @@ int main(int argc, char **argv)
 		case 'p':
 			loop_tests = 1;
 			break;
+		case 'q':
+			quiet = true;
+			break;
 		case 'h':
 			print_usage();
 			print_help();
@@ -967,18 +1072,25 @@ int main(int argc, char **argv)
 		}
 	};
 
-	if (firmware_analysis)
-		analyze_firmware(firmware_analysis);
+	if (!quiet) {
+		printf("ipaccess-config (C) 2009-2010 by Harald Welte and others\n");
+		printf("This is FREE SOFTWARE with ABSOLUTELY NO WARRANTY\n\n");
+	}
 
-	if (optind >= argc) {
-		/* only warn if we have not done anything else */
-		if (!firmware_analysis)
-			fprintf(stderr, "you have to specify the IP address of the BTS. Use --help for more information\n");
+	if (firmware_analysis) {
+		analyze_firmware(firmware_analysis);
+		if (argc == optind) /* Nothing more to do, exit successfully */
+			exit(EXIT_SUCCESS);
+	}
+	if (argc - optind != 1) {
+		fprintf(stderr, "you have to specify the IP address of the BTS. Use --help for more information\n");
 		exit(2);
 	}
+	bts_ip = argv[optind++];
+
 	libosmo_abis_init(tall_ctx_config);
 
-	bsc_gsmnet = bsc_network_init(tall_bsc_ctx, 1, 1, NULL);
+	bsc_gsmnet = gsm_network_init(tall_ctx_config);
 	if (!bsc_gsmnet)
 		exit(1);
 
@@ -989,23 +1101,24 @@ int main(int argc, char **argv)
 	gsm_bts_trx_alloc(bts);
 	gsm_bts_trx_alloc(bts);
 	bts->oml_tei = stream_id;
-	
+
 	osmo_signal_register_handler(SS_NM, nm_sig_cb, NULL);
 	osmo_signal_register_handler(SS_IPAC_NWL, nwl_sig_cb, NULL);
 
 	ipac_nwl_init();
 
-	printf("Trying to connect to ip.access BTS ...\n");
+	if (!quiet)
+		printf("Trying to connect to ip.access BTS %s...\n", bts_ip);
 
 	memset(&sin, 0, sizeof(sin));
 	sin.sin_family = AF_INET;
-	inet_aton(argv[optind], &sin.sin_addr);
+	inet_aton(bts_ip, &sin.sin_addr);
 	rc = ia_config_connect(bts, &sin);
 	if (rc < 0) {
 		perror("Error connecting to the BTS");
 		exit(1);
 	}
-	
+
 	bts->oml_link->ts->sign.delay = 10;
 	bts->c0->rsl_link->ts->sign.delay = 10;
 	while (1) {
@@ -1017,3 +1130,20 @@ int main(int argc, char **argv)
 	exit(0);
 }
 
+/* Stub */
+int osmo_bsc_sigtran_send(struct gsm_subscriber_connection *conn, struct msgb *msg)
+{
+	return 0;
+}
+
+/* Stub */
+int osmo_bsc_sigtran_open_conn(struct gsm_subscriber_connection *conn, struct msgb *msg)
+{
+	return 0;
+}
+
+/* Stub */
+int rsl_chan_ms_power_ctrl(struct gsm_lchan *lchan)
+{
+	return 0;
+}
