@@ -19,56 +19,205 @@
  */
 
 #include <osmocom/ctrl/control_cmd.h>
-#include <openbsc/debug.h>
-#include <openbsc/gsm_data.h>
-#include <openbsc/osmo_bsc.h>
-#include <openbsc/osmo_bsc_rf.h>
-#include <openbsc/bsc_msc_data.h>
-#include <openbsc/signal.h>
-#include <openbsc/gsm_04_80.h>
+#include <osmocom/bsc/ctrl.h>
+#include <osmocom/bsc/debug.h>
+#include <osmocom/bsc/gsm_data.h>
+#include <osmocom/bsc/osmo_bsc.h>
+#include <osmocom/bsc/osmo_bsc_rf.h>
+#include <osmocom/bsc/bsc_msc_data.h>
+#include <osmocom/bsc/signal.h>
+#include <osmocom/bsc/gsm_04_80.h>
 
 #include <osmocom/core/linuxlist.h>
 #include <osmocom/core/signal.h>
-#include <osmocom/core/talloc.h>
+
+#include <osmocom/ctrl/control_if.h>
+
+#include <osmocom/gsm/protocol/ipaccess.h>
+#include <osmocom/gsm/ipa.h>
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <time.h>
 #include <unistd.h>
 
-void osmo_bsc_send_trap(struct ctrl_cmd *cmd, struct bsc_msc_connection *msc_con)
+/* Obtain SS7 application server currently handling given MSC (DPC) */
+static struct osmo_ss7_as *msc_get_ss7_as(struct bsc_msc_data *msc)
+{
+	struct osmo_ss7_route *rt;
+	struct osmo_ss7_instance *ss7 = osmo_sccp_get_ss7(msc->a.sccp);
+	rt = osmo_ss7_route_lookup(ss7, msc->a.msc_addr.pc);
+	if (!rt)
+		return NULL;
+	return rt->dest.as;
+}
+
+static int _ss7_as_send(struct osmo_ss7_as *as, struct msgb *msg)
+{
+	struct osmo_ss7_asp *asp;
+	unsigned int i;
+
+	/* FIXME: unify with xua_as_transmit_msg() and perform proper ASP lookup */
+	for (i = 0; i < ARRAY_SIZE(as->cfg.asps); i++) {
+		asp = as->cfg.asps[i];
+		if (!asp)
+			continue;
+		/* FIXME: deal with multiple ASPs per AS */
+		return osmo_ss7_asp_send(asp, msg);
+	}
+	msgb_free(msg);
+	return -1;
+}
+
+int bsc_sccplite_msc_send(struct bsc_msc_data *msc, struct msgb *msg)
+{
+	struct osmo_ss7_as *as;
+
+	as = msc_get_ss7_as(msc);
+	if (!as) {
+		msgb_free(msg);
+		return -1;
+	}
+
+	/* don't attempt to send CTRL on a non-SCCPlite AS */
+	if (as->cfg.proto != OSMO_SS7_ASP_PROT_IPA)
+		return 0;
+
+	return _ss7_as_send(as, msg);
+}
+
+/* Encode a CTRL command and send it to the given ASP
+ * \param[in] asp ASP through which we shall send the encoded message
+ * \param[in] cmd decoded CTRL command to be encoded and sent. Ownership is *NOT*
+ * 		  transferred, to permit caller to send the same CMD to several ASPs.
+ * 		  Caller must hence free 'cmd' itself.
+ * \returns 0 on success; negative on error */
+static int sccplite_asp_ctrl_cmd_send(struct osmo_ss7_asp *asp, struct ctrl_cmd *cmd)
+{
+	/* this is basically like libosmoctrl:ctrl_cmd_send(), not for a dedicated
+	 * CTRL connection but for the CTRL piggy-back on the IPA/SCCPlite link */
+	struct msgb *msg;
+
+	/* don't attempt to send CTRL on a non-SCCPlite ASP */
+	if (asp->cfg.proto != OSMO_SS7_ASP_PROT_IPA)
+		return 0;
+
+	msg = ctrl_cmd_make(cmd);
+	if (!msg)
+		return -1;
+
+	ipa_prepend_header_ext(msg, IPAC_PROTO_EXT_CTRL);
+	ipa_prepend_header(msg, IPAC_PROTO_OSMO);
+
+	return osmo_ss7_asp_send(asp, msg);
+}
+
+/* Ownership of 'cmd' is *NOT* transferred, to permit caller to send the same CMD to several ASPs.
+ * Caller must hence free 'cmd' itself. */
+static int sccplite_msc_ctrl_cmd_send(struct bsc_msc_data *msc, struct ctrl_cmd *cmd)
+{
+	struct msgb *msg;
+
+	msg = ctrl_cmd_make(cmd);
+	if (!msg)
+		return -1;
+
+	ipa_prepend_header_ext(msg, IPAC_PROTO_EXT_CTRL);
+	ipa_prepend_header(msg, IPAC_PROTO_OSMO);
+
+	return bsc_sccplite_msc_send(msc, msg);
+}
+
+/* receive + process a CTRL command from the piggy-back on the IPA/SCCPlite link.
+ * Transfers msg ownership. */
+int bsc_sccplite_rx_ctrl(struct osmo_ss7_asp *asp, struct msgb *msg)
+{
+	struct ctrl_cmd *cmd;
+	bool parse_failed;
+	int rc;
+
+	/* caller has already ensured ipaccess_head + ipaccess_head_ext */
+	OSMO_ASSERT(msg->l2h);
+
+	/* prase raw (ASCII) CTRL command into ctrl_cmd */
+	cmd = ctrl_cmd_parse3(asp, msg, &parse_failed);
+	OSMO_ASSERT(cmd);
+	msgb_free(msg);
+	if (cmd->type == CTRL_TYPE_ERROR && parse_failed)
+		goto send_reply;
+
+	/* handle the CTRL command */
+	ctrl_cmd_handle(bsc_gsmnet->ctrl, cmd, bsc_gsmnet);
+
+send_reply:
+	rc = sccplite_asp_ctrl_cmd_send(asp, cmd);
+	talloc_free(cmd);
+	return rc;
+}
+
+
+void osmo_bsc_send_trap(struct ctrl_cmd *cmd, struct bsc_msc_data *msc_data)
 {
 	struct ctrl_cmd *trap;
 	struct ctrl_handle *ctrl;
-	struct bsc_msc_data *msc_data;
 
-	msc_data = (struct bsc_msc_data *) msc_con->write_queue.bfd.data;
 	ctrl = msc_data->network->ctrl;
 
 	trap = ctrl_cmd_trap(cmd);
 	if (!trap) {
+
 		LOGP(DCTRL, LOGL_ERROR, "Failed to create trap.\n");
 		return;
 	}
 
 	ctrl_cmd_send_to_all(ctrl, trap);
-	ctrl_cmd_send(&msc_con->write_queue, trap);
+	sccplite_msc_ctrl_cmd_send(msc_data, trap);
 
 	talloc_free(trap);
 }
 
-CTRL_CMD_DEFINE_RO(msc_connection_status, "msc_connection_status");
-static int msc_connection_status = 0;
-
+CTRL_CMD_DEFINE_RO(msc_connection_status, "connection_status");
 static int get_msc_connection_status(struct ctrl_cmd *cmd, void *data)
 {
-	if (msc_connection_status)
+	struct bsc_msc_data *msc = (struct bsc_msc_data *)cmd->node;
+	struct osmo_ss7_as *as;
+	const char *as_state_name;
+
+	if (msc == NULL) {
+		cmd->reply = "msc not found";
+		return CTRL_CMD_ERROR;
+	}
+	as = msc_get_ss7_as(msc);
+	if (!as) {
+		cmd->reply = "AS not found for MSC";
+		return CTRL_CMD_ERROR;
+	}
+
+	as_state_name = osmo_fsm_inst_state_name(as->fi);
+	if (!strcmp(as_state_name, "AS_ACTIVE"))
 		cmd->reply = "connected";
 	else
 		cmd->reply = "disconnected";
 	return CTRL_CMD_REPLY;
 }
 
+/* Backwards compat. */
+CTRL_CMD_DEFINE_RO(msc0_connection_status, "msc_connection_status");
+static int msc_connection_status = 0; /* XXX unused */
+
+static int get_msc0_connection_status(struct ctrl_cmd *cmd, void *data)
+{
+	struct bsc_msc_data *msc = osmo_msc_data_find(bsc_gsmnet, 0);
+	void *old_node = cmd->node;
+	int rc;
+
+	cmd->node = msc;
+	rc = get_msc_connection_status(cmd, data);
+	cmd->node = old_node;
+
+	return rc;
+}
+
 static int msc_connection_status_trap_cb(unsigned int subsys, unsigned int signal, void *handler_data, void *signal_data)
 {
 	struct ctrl_cmd *cmd;
@@ -93,7 +242,7 @@ static int msc_connection_status_trap_cb(unsigned int subsys, unsigned int signa
 	cmd->id = "0";
 	cmd->variable = "msc_connection_status";
 
-	get_msc_connection_status(cmd, NULL);
+	get_msc0_connection_status(cmd, NULL);
 
 	ctrl_cmd_send_to_all(gsmnet->ctrl, cmd);
 
@@ -163,12 +312,12 @@ static int bts_connection_status_trap_cb(unsigned int subsys, unsigned int signa
 
 static int get_bts_loc(struct ctrl_cmd *cmd, void *data);
 
-static void generate_location_state_trap(struct gsm_bts *bts, struct bsc_msc_connection *msc_con)
+static void generate_location_state_trap(struct gsm_bts *bts, struct bsc_msc_data *msc)
 {
 	struct ctrl_cmd *cmd;
 	const char *oper, *admin, *policy;
 
-	cmd = ctrl_cmd_create(msc_con, CTRL_TYPE_TRAP);
+	cmd = ctrl_cmd_create(msc, CTRL_TYPE_TRAP);
 	if (!cmd) {
 		LOGP(DCTRL, LOGL_ERROR, "Failed to create TRAP command.\n");
 		return;
@@ -186,12 +335,13 @@ static void generate_location_state_trap(struct gsm_bts *bts, struct bsc_msc_con
 	policy = osmo_bsc_rf_get_policy_name(osmo_bsc_rf_get_policy_by_bts(bts));
 
 	cmd->reply = talloc_asprintf_append(cmd->reply,
-				",%s,%s,%s,%d,%d",
+				",%s,%s,%s,%s,%s",
 				oper, admin, policy,
-				bts->network->country_code,
-				bts->network->network_code);
+				osmo_mcc_name(bts->network->plmn.mcc),
+				osmo_mnc_name(bts->network->plmn.mnc,
+					      bts->network->plmn.mnc_3_digits));
 
-	osmo_bsc_send_trap(cmd, msc_con);
+	osmo_bsc_send_trap(cmd, msc);
 	talloc_free(cmd);
 }
 
@@ -200,7 +350,7 @@ void bsc_gen_location_state_trap(struct gsm_bts *bts)
 	struct bsc_msc_data *msc;
 
 	llist_for_each_entry(msc, &bts->network->bsc_data->mscs, entry)
-		generate_location_state_trap(bts, msc->msc_con);
+		generate_location_state_trap(bts, msc);
 }
 
 static int location_equal(struct bts_location *a, struct bts_location *b)
@@ -385,8 +535,9 @@ static int get_net_timezone(struct ctrl_cmd *cmd, void *data)
 static int set_net_timezone(struct ctrl_cmd *cmd, void *data)
 {
 	char *saveptr, *hourstr, *minstr, *dststr, *tmp = 0;
-	int override;
+	int override = 0;
 	struct gsm_network *net = (struct gsm_network*)cmd->node;
+	struct gsm_tz *tz = &net->tz;
 
 	tmp = talloc_strdup(cmd, cmd->value);
 	if (!tmp)
@@ -396,19 +547,17 @@ static int set_net_timezone(struct ctrl_cmd *cmd, void *data)
 	minstr = strtok_r(NULL, ",", &saveptr);
 	dststr = strtok_r(NULL, ",", &saveptr);
 
-	override = 0;
-
-	if (hourstr != NULL)
+	if (hourstr != NULL) {
 		override = strcasecmp(hourstr, "off") != 0;
+		if (override) {
+			tz->hr  = atol(hourstr);
+			tz->mn  = minstr ? atol(minstr) : 0;
+			tz->dst = dststr ? atol(dststr) : 0;
+		}
+	}
 
-	struct gsm_tz *tz = &net->tz;
 	tz->override = override;
 
-	if (override) {
-		tz->hr  = hourstr ? atol(hourstr) : 0;
-		tz->mn  = minstr ? atol(minstr) : 0;
-		tz->dst = dststr ? atol(dststr) : 0;
-	}
 
 	talloc_free(tmp);
 	tmp = NULL;
@@ -466,13 +615,7 @@ err:
 	return 1;
 }
 
-CTRL_CMD_DEFINE(net_notification, "notification");
-static int get_net_notification(struct ctrl_cmd *cmd, void *data)
-{
-	cmd->reply = "There is nothing to read";
-	return CTRL_CMD_ERROR;
-}
-
+CTRL_CMD_DEFINE_WO_NOVRF(net_notification, "notification");
 static int set_net_notification(struct ctrl_cmd *cmd, void *data)
 {
 	struct ctrl_cmd *trap;
@@ -502,18 +645,7 @@ handled:
 	return CTRL_CMD_HANDLED;
 }
 
-static int verify_net_notification(struct ctrl_cmd *cmd, const char *value, void *data)
-{
-	return 0;
-}
-
-CTRL_CMD_DEFINE(net_inform_msc, "inform-msc-v1");
-static int get_net_inform_msc(struct ctrl_cmd *cmd, void *data)
-{
-	cmd->reply = "There is nothing to read";
-	return CTRL_CMD_ERROR;
-}
-
+CTRL_CMD_DEFINE_WO_NOVRF(net_inform_msc, "inform-msc-v1");
 static int set_net_inform_msc(struct ctrl_cmd *cmd, void *data)
 {
 	struct gsm_network *net;
@@ -532,7 +664,7 @@ static int set_net_inform_msc(struct ctrl_cmd *cmd, void *data)
 		trap->id = "0";
 		trap->variable = "inform-msc-v1";
 		trap->reply = talloc_strdup(trap, cmd->value);
-		ctrl_cmd_send(&msc->msc_con->write_queue, trap);
+		sccplite_msc_ctrl_cmd_send(msc, trap);
 		talloc_free(trap);
 	}
 
@@ -540,18 +672,7 @@ static int set_net_inform_msc(struct ctrl_cmd *cmd, void *data)
 	return CTRL_CMD_HANDLED;
 }
 
-static int verify_net_inform_msc(struct ctrl_cmd *cmd, const char *value, void *data)
-{
-	return 0;
-}
-
-CTRL_CMD_DEFINE(net_ussd_notify, "ussd-notify-v1");
-static int get_net_ussd_notify(struct ctrl_cmd *cmd, void *data)
-{
-	cmd->reply = "There is nothing to read";
-	return CTRL_CMD_ERROR;
-}
-
+CTRL_CMD_DEFINE_WO(net_ussd_notify, "ussd-notify-v1");
 static int set_net_ussd_notify(struct ctrl_cmd *cmd, void *data)
 {
 	struct gsm_subscriber_connection *conn;
@@ -577,10 +698,7 @@ static int set_net_ussd_notify(struct ctrl_cmd *cmd, void *data)
 
 	net = cmd->node;
 	llist_for_each_entry(conn, &net->subscr_conns, entry) {
-		if (!conn->sccp_con)
-			continue;
-
-		if (conn->sccp_con->cic != cic)
+		if (conn->user_plane.msc_assigned_cic != cic)
 			continue;
 
 		/*
@@ -634,7 +752,7 @@ static int msc_signal_handler(unsigned int subsys, unsigned int signal,
 
 	net = msc->data->network;
 	llist_for_each_entry(bts, &net->bts_list, list)
-		generate_location_state_trap(bts, msc->data->msc_con);	
+		generate_location_state_trap(bts, msc->data);
 
 	return 0;
 }
@@ -652,7 +770,10 @@ int bsc_ctrl_cmds_install(struct gsm_network *net)
 	rc = ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_net_timezone);
 	if (rc)
 		goto end;
-	rc = ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_msc_connection_status);
+	rc = ctrl_cmd_install(CTRL_NODE_MSC, &cmd_msc_connection_status);
+	if (rc)
+		goto end;
+	rc = ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_msc0_connection_status);
 	if (rc)
 		goto end;
 	rc = osmo_signal_register_handler(SS_MSC, &msc_connection_status_trap_cb, net);
