@@ -1,6 +1,7 @@
 /* Osmo BSC VTY Configuration */
 /* (C) 2009-2015 by Holger Hans Peter Freyther
  * (C) 2009-2014 by On-Waves
+ * (C) 2018 by Harald Welte <laforge@gnumonks.org>
  * All Rights Reserved
  *
  * This program is free software; you can redistribute it and/or modify
@@ -18,25 +19,26 @@
  *
  */
 
-#include <openbsc/gsm_data.h>
-#include <openbsc/osmo_bsc.h>
-#include <openbsc/bsc_msc_data.h>
-#include <openbsc/vty.h>
-#include <openbsc/bsc_subscriber.h>
-#include <openbsc/debug.h>
-#include <openbsc/bsc_msg_filter.h>
+#include <osmocom/bsc/gsm_data.h>
+#include <osmocom/bsc/osmo_bsc.h>
+#include <osmocom/bsc/bsc_msc_data.h>
+#include <osmocom/bsc/vty.h>
+#include <osmocom/bsc/bsc_subscriber.h>
+#include <osmocom/bsc/debug.h>
+#include <osmocom/bsc/bsc_msg_filter.h>
+#include <osmocom/bsc/osmux.h>
 
 #include <osmocom/core/talloc.h>
+#include <osmocom/gsm/gsm48.h>
 #include <osmocom/vty/logging.h>
-#include <osmocom/sccp/sccp_types.h>
+#include <osmocom/mgcp_client/mgcp_client.h>
+
 
 #include <time.h>
 
 
 #define IPA_STR "IP.ACCESS specific\n"
 
-extern struct gsm_network *bsc_gsmnet;
-
 static struct osmo_bsc_data *osmo_bsc_data(struct vty *vty)
 {
 	return bsc_gsmnet->bsc_data;
@@ -99,24 +101,22 @@ static void write_msc_amr_options(struct vty *vty, struct bsc_msc_data *msc)
 	WRITE_AMR(vty, msc, "5_15k", m5_15);
 	WRITE_AMR(vty, msc, "4_75k", m4_75);
 #undef WRITE_AMR
+
+	if (msc->amr_octet_aligned)
+		vty_out(vty, " amr-payload octet-aligned%s", VTY_NEWLINE);
+	else
+		vty_out(vty, " amr-payload bandwith-efficient%s", VTY_NEWLINE);
 }
 
 static void write_msc(struct vty *vty, struct bsc_msc_data *msc)
 {
-	struct bsc_msc_dest *dest;
-
 	vty_out(vty, "msc %d%s", msc->nr, VTY_NEWLINE);
-	if (msc->bsc_token)
-		vty_out(vty, " token %s%s", msc->bsc_token, VTY_NEWLINE);
-	if (msc->bsc_key_present)
-		vty_out(vty, " auth-key %s%s",
-			osmo_hexdump(msc->bsc_key, sizeof(msc->bsc_key)), VTY_NEWLINE);
-	if (msc->core_mnc != -1)
-		vty_out(vty, " core-mobile-network-code %d%s",
-			msc->core_mnc, VTY_NEWLINE);
-	if (msc->core_mcc != -1)
-		vty_out(vty, " core-mobile-country-code %d%s",
-			msc->core_mcc, VTY_NEWLINE);
+	if (msc->core_plmn.mnc != GSM_MCC_MNC_INVALID)
+		vty_out(vty, " core-mobile-network-code %s%s",
+			osmo_mnc_name(msc->core_plmn.mnc, msc->core_plmn.mnc_3_digits), VTY_NEWLINE);
+	if (msc->core_plmn.mcc != GSM_MCC_MNC_INVALID)
+		vty_out(vty, " core-mobile-country-code %s%s",
+			osmo_mcc_name(msc->core_plmn.mcc), VTY_NEWLINE);
 	if (msc->core_lac != -1)
 		vty_out(vty, " core-location-area-code %d%s",
 			msc->core_lac, VTY_NEWLINE);
@@ -125,17 +125,6 @@ static void write_msc(struct vty *vty, struct bsc_msc_data *msc)
 			msc->core_ci, VTY_NEWLINE);
 	vty_out(vty, " ip.access rtp-base %d%s", msc->rtp_base, VTY_NEWLINE);
 
-	if (msc->ping_timeout == -1)
-		vty_out(vty, " no timeout-ping%s", VTY_NEWLINE);
-	else {
-		vty_out(vty, " timeout-ping %d%s", msc->ping_timeout, VTY_NEWLINE);
-		vty_out(vty, " timeout-pong %d%s", msc->pong_timeout, VTY_NEWLINE);
-		if (msc->advanced_ping)
-			vty_out(vty, " timeout-ping advanced%s", VTY_NEWLINE);
-		else
-			vty_out(vty, " no timeout-ping advanced%s", VTY_NEWLINE);
-	}
-
 	if (msc->ussd_welcome_txt)
 		vty_out(vty, " bsc-welcome-text %s%s", msc->ussd_welcome_txt, VTY_NEWLINE);
 	else
@@ -168,10 +157,6 @@ static void write_msc(struct vty *vty, struct bsc_msc_data *msc)
 
 	}
 
-	llist_for_each_entry(dest, &msc->dests, list)
-		vty_out(vty, " dest %s %d %d%s", dest->ip, dest->port,
-			dest->dscp, VTY_NEWLINE);
-
 	vty_out(vty, " type %s%s", msc->type == MSC_CON_TYPE_NORMAL ?
 					"normal" : "local", VTY_NEWLINE);
 	vty_out(vty, " allow-emergency %s%s", msc->allow_emerg ?
@@ -187,13 +172,37 @@ static void write_msc(struct vty *vty, struct bsc_msc_data *msc)
 	write_msc_amr_options(vty, msc);
 
 	/* write sccp connection configuration */
-	/* FIXME: This can not work, as we manipulate the address,
-	 * we can not expect to find it in the addressbok. We should
-	 * store the string names instead. */
-	vty_out(vty, " bsc-addr %s%s",
-		osmo_sccp_name_by_addr(&msc->a.bsc_addr), VTY_NEWLINE);
-	vty_out(vty, " msc-addr %s%s",
-		osmo_sccp_name_by_addr(&msc->a.msc_addr), VTY_NEWLINE);
+	if (msc->a.bsc_addr_name) {
+		vty_out(vty, " bsc-addr %s%s",
+			msc->a.bsc_addr_name, VTY_NEWLINE);
+	}
+	if (msc->a.msc_addr_name) {
+		vty_out(vty, " msc-addr %s%s",
+			msc->a.msc_addr_name, VTY_NEWLINE);
+	}
+	vty_out(vty, " asp-protocol %s%s", osmo_ss7_asp_protocol_name(msc->a.asp_proto), VTY_NEWLINE);
+	vty_out(vty, " lcls-mode %s%s", get_value_string(bsc_lcls_mode_names, msc->lcls_mode),
+		VTY_NEWLINE);
+
+	if (msc->lcls_codec_mismatch_allow)
+		vty_out(vty, " lcls-codec-mismatch allowed%s", VTY_NEWLINE);
+	else
+		vty_out(vty, " lcls-codec-mismatch forbidden%s", VTY_NEWLINE);
+
+	/* write MGW configuration */
+	mgcp_client_config_write(vty, " ");
+
+	if (msc->x_osmo_ign_configured) {
+		if (!msc->x_osmo_ign)
+			vty_out(vty, " no mgw x-osmo-ign%s", VTY_NEWLINE);
+		else
+			vty_out(vty, " mgw x-osmo-ign call-id%s", VTY_NEWLINE);
+	}
+
+	if (msc->use_osmux != OSMUX_USAGE_OFF) {
+		vty_out(vty, " osmux %s%s", msc->use_osmux == OSMUX_USAGE_ON ? "on" : "only",
+			VTY_NEWLINE);
+	}
 }
 
 static int config_write_msc(struct vty *vty)
@@ -230,41 +239,8 @@ static int config_write_bsc(struct vty *vty)
 	if (bsc->acc_lst_name)
 		vty_out(vty, " access-list-name %s%s", bsc->acc_lst_name, VTY_NEWLINE);
 
-	return CMD_SUCCESS;
-}
+	bsc_msg_acc_lst_write(vty);
 
-DEFUN(cfg_net_bsc_token,
-      cfg_net_bsc_token_cmd,
-      "token TOKEN",
-      "A token for the BSC to be sent to the MSC\n" "A token\n")
-{
-	struct bsc_msc_data *data = bsc_msc_data(vty);
-
-	osmo_talloc_replace_string(osmo_bsc_data(vty), &data->bsc_token, argv[0]);
-	return CMD_SUCCESS;
-}
-
-DEFUN(cfg_net_bsc_key,
-      cfg_net_bsc_key_cmd,
-      "auth-key KEY",
-      "Authentication (secret) key configuration\n"
-      "Security key\n")
-{
-	struct bsc_msc_data *data = bsc_msc_data(vty);
-
-	osmo_hexparse(argv[0], data->bsc_key, sizeof(data->bsc_key));
-	data->bsc_key_present = 1;
-	return CMD_SUCCESS;
-}
-
-DEFUN(cfg_net_no_bsc_key, cfg_net_bsc_no_key_cmd,
-      "no auth-key",
-      NO_STR "Authentication (secret) key configuration\n")
-{
-	struct bsc_msc_data *data = bsc_msc_data(vty);
-
-	memset(data->bsc_key, 0, sizeof(data->bsc_key));
-	data->bsc_key_present = 0;
 	return CMD_SUCCESS;
 }
 
@@ -274,7 +250,15 @@ DEFUN(cfg_net_bsc_ncc,
       "Use this network code for the core network\n" "MNC value\n")
 {
 	struct bsc_msc_data *data = bsc_msc_data(vty);
-	data->core_mnc = atoi(argv[0]);
+	uint16_t mnc;
+	bool mnc_3_digits;
+
+	if (osmo_mnc_from_str(argv[0], &mnc, &mnc_3_digits)) {
+		vty_out(vty, "%% Error decoding MNC: %s%s", argv[0], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+	data->core_plmn.mnc = mnc;
+	data->core_plmn.mnc_3_digits = mnc_3_digits;
 	return CMD_SUCCESS;
 }
 
@@ -283,8 +267,13 @@ DEFUN(cfg_net_bsc_mcc,
       "core-mobile-country-code <1-999>",
       "Use this country code for the core network\n" "MCC value\n")
 {
+	uint16_t mcc;
 	struct bsc_msc_data *data = bsc_msc_data(vty);
-	data->core_mcc = atoi(argv[0]);
+	if (osmo_mcc_from_str(argv[0], &mcc)) {
+		vty_out(vty, "%% Error decoding MCC: %s%s", argv[0], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+	data->core_plmn.mcc = mcc;
 	return CMD_SUCCESS;
 }
 
@@ -327,11 +316,8 @@ DEFUN(cfg_net_bsc_codec_list,
       "List of audio codecs, e.g. fr3 fr1 hr3\n")
 {
 	struct bsc_msc_data *data = bsc_msc_data(vty);
-	int saw_fr, saw_hr;
 	int i;
 
-	saw_fr = saw_hr = 0;
-
 	/* free the old list... if it exists */
 	if (data->audio_support) {
 		talloc_free(data->audio_support);
@@ -357,19 +343,10 @@ DEFUN(cfg_net_bsc_codec_list,
 				struct gsm_audio_support);
 		data->audio_support[i]->ver = atoi(argv[i] + 2);
 
-		if (strncmp("hr", argv[i], 2) == 0) {
+		if (strncmp("hr", argv[i], 2) == 0)
 			data->audio_support[i]->hr = 1;
-			saw_hr = 1;
-		} else if (strncmp("fr", argv[i], 2) == 0) {
+		else if (strncmp("fr", argv[i], 2) == 0)
 			data->audio_support[i]->hr = 0;
-			saw_fr = 1;
-		}
-
-		if (saw_hr && saw_fr) {
-			vty_out(vty, "Can not have full-rate and half-rate codec.%s",
-					VTY_NEWLINE);
-			return CMD_ERR_INCOMPLETE;
-		}
 	}
 
 	return CMD_SUCCESS;
@@ -380,116 +357,6 @@ error:
 	return CMD_ERR_INCOMPLETE;
 }
 
-DEFUN(cfg_net_msc_dest,
-      cfg_net_msc_dest_cmd,
-      "dest A.B.C.D <1-65000> <0-255>",
-      "Add a destination to a MUX/MSC\n"
-      "IP Address\n" "Port\n" "DSCP\n")
-{
-	struct bsc_msc_dest *dest;
-	struct bsc_msc_data *data = bsc_msc_data(vty);
-
-	dest = talloc_zero(osmo_bsc_data(vty), struct bsc_msc_dest);
-	if (!dest) {
-		vty_out(vty, "%%Failed to create structure.%s", VTY_NEWLINE);
-		return CMD_WARNING;
-	}
-
-	dest->ip = talloc_strdup(dest, argv[0]);
-	if (!dest->ip) {
-		vty_out(vty, "%%Failed to copy dest ip.%s", VTY_NEWLINE);
-		talloc_free(dest);
-		return CMD_WARNING;
-	}
-
-	dest->port = atoi(argv[1]);
-	dest->dscp = atoi(argv[2]);
-	llist_add_tail(&dest->list, &data->dests);
-	return CMD_SUCCESS;
-}
-
-DEFUN(cfg_net_msc_no_dest,
-      cfg_net_msc_no_dest_cmd,
-      "no dest A.B.C.D <1-65000> <0-255>",
-      NO_STR "Remove a destination to a MUX/MSC\n"
-      "IP Address\n" "Port\n" "DSCP\n")
-{
-	struct bsc_msc_dest *dest, *tmp;
-	struct bsc_msc_data *data = bsc_msc_data(vty);
-
-	int port = atoi(argv[1]);
-	int dscp = atoi(argv[2]);
-
-	llist_for_each_entry_safe(dest, tmp, &data->dests, list) {
-		if (port != dest->port || dscp != dest->dscp
-		    || strcmp(dest->ip, argv[0]) != 0)
-			continue;
-
-		llist_del(&dest->list);
-		talloc_free(dest);
-	}
-
-	return CMD_SUCCESS;
-}
-
-DEFUN(cfg_net_msc_no_ping_time,
-      cfg_net_msc_no_ping_time_cmd,
-      "no timeout-ping",
-      NO_STR "Disable the ping/pong handling on A-link\n")
-{
-	struct bsc_msc_data *data = bsc_msc_data(vty);
-	data->ping_timeout = -1;
-	return CMD_SUCCESS;
-}
-
-DEFUN(cfg_net_msc_ping_time,
-      cfg_net_msc_ping_time_cmd,
-      "timeout-ping <1-2147483647>",
-      "Set the PING interval, negative for not sending PING\n"
-      "Timeout in seconds\n")
-{
-	struct bsc_msc_data *data = bsc_msc_data(vty);
-	data->ping_timeout = atoi(argv[0]);
-	return CMD_SUCCESS;
-}
-
-DEFUN(cfg_net_msc_pong_time,
-      cfg_net_msc_pong_time_cmd,
-      "timeout-pong <1-2147483647>",
-      "Set the time to wait for a PONG\n" "Timeout in seconds\n")
-{
-	struct bsc_msc_data *data = bsc_msc_data(vty);
-	data->pong_timeout = atoi(argv[0]);
-	return CMD_SUCCESS;
-}
-
-DEFUN(cfg_net_msc_advanced_ping,
-      cfg_net_msc_advanced_ping_cmd,
-      "timeout-ping advanced",
-      "Ping timeout handling\nEnable advanced mode during SCCP\n")
-{
-	struct bsc_msc_data *data = bsc_msc_data(vty);
-
-	if (data->ping_timeout == -1) {
-		vty_out(vty, "%%ping handling is disabled. Enable it first.%s",
-			VTY_NEWLINE);
-		return CMD_WARNING;
-	}
-
-	data->advanced_ping = 1;
-	return CMD_SUCCESS;
-}
-
-DEFUN(cfg_no_net_msc_advanced_ping,
-      cfg_no_net_msc_advanced_ping_cmd,
-      "no timeout-ping advanced",
-      NO_STR "Ping timeout handling\nEnable advanced mode during SCCP\n")
-{
-	struct bsc_msc_data *data = bsc_msc_data(vty);
-	data->advanced_ping = 0;
-	return CMD_SUCCESS;
-}
-
 DEFUN(cfg_net_msc_welcome_ussd,
       cfg_net_msc_welcome_ussd_cmd,
       "bsc-welcome-text .TEXT",
@@ -700,14 +567,14 @@ DEFUN(cfg_msc_no_acc_lst_name,
 static void enforce_standard_ssn(struct vty *vty, struct osmo_sccp_addr *addr)
 {
 	if (addr->presence & OSMO_SCCP_ADDR_T_SSN) {
-		if (addr->ssn != SCCP_SSN_BSSAP)
+		if (addr->ssn != OSMO_SCCP_SSN_BSSAP)
 			vty_out(vty,
-				"setting an SSN (%u) different from the standard (%u) is not allowd, will use standard SSN for address: %s%s",
-				addr->ssn, SCCP_SSN_BSSAP, osmo_sccp_addr_dump(addr), VTY_NEWLINE);
+				"setting an SSN (%u) different from the standard (%u) is not allowed, will use standard SSN for address: %s%s",
+				addr->ssn, OSMO_SCCP_SSN_BSSAP, osmo_sccp_addr_dump(addr), VTY_NEWLINE);
 	}
 
 	addr->presence |= OSMO_SCCP_ADDR_T_SSN;
-	addr->ssn = SCCP_SSN_BSSAP;
+	addr->ssn = OSMO_SCCP_SSN_BSSAP;
 }
 
 DEFUN(cfg_msc_cs7_bsc_addr,
@@ -721,24 +588,24 @@ DEFUN(cfg_msc_cs7_bsc_addr,
 
 	ss7 = osmo_sccp_addr_by_name(&msc->a.bsc_addr, bsc_addr_name);
 	if (!ss7) {
-		vty_out(vty, "No sccp address %s found%s", bsc_addr_name,
-			VTY_NEWLINE);
-		return CMD_WARNING;
+		vty_out(vty, "Error: No such SCCP addressbook entry: '%s'%s", bsc_addr_name, VTY_NEWLINE);
+		return CMD_ERR_INCOMPLETE;
 	}
 
 	/* Prevent mixing addresses from different CS7/SS7 instances */
 	if (msc->a.cs7_instance_valid) {
 		if (msc->a.cs7_instance != ss7->cfg.id) {
 			vty_out(vty,
-				"SCCP address %s from different CS7 instance%s",
+				"Error: SCCP addressbook entry from mismatching CS7 instance: '%s'%s",
 				bsc_addr_name, VTY_NEWLINE);
-			return CMD_WARNING;
+			return CMD_ERR_INCOMPLETE;
 		}
 	}
 
 	msc->a.cs7_instance = ss7->cfg.id;
 	msc->a.cs7_instance_valid = true;
 	enforce_standard_ssn(vty, &msc->a.bsc_addr);
+	msc->a.bsc_addr_name = talloc_strdup(msc, bsc_addr_name);
 	return CMD_SUCCESS;
 }
 
@@ -753,31 +620,122 @@ DEFUN(cfg_msc_cs7_msc_addr,
 
 	ss7 = osmo_sccp_addr_by_name(&msc->a.msc_addr, msc_addr_name);
 	if (!ss7) {
-		vty_out(vty, "No sccp address %s found%s", msc_addr_name,
-			VTY_NEWLINE);
-		return CMD_WARNING;
+		vty_out(vty, "Error: No such SCCP addressbook entry: '%s'%s", msc_addr_name, VTY_NEWLINE);
+		return CMD_ERR_INCOMPLETE;
 	}
 
 	/* Prevent mixing addresses from different CS7/SS7 instances */
 	if (msc->a.cs7_instance_valid) {
 		if (msc->a.cs7_instance != ss7->cfg.id) {
 			vty_out(vty,
-				"SCCP address %s from different CS7 instance%s",
+				"Error: SCCP addressbook entry from mismatching CS7 instance: '%s'%s",
 				msc_addr_name, VTY_NEWLINE);
-			return CMD_WARNING;
+			return CMD_ERR_INCOMPLETE;
 		}
 	}
 
 	msc->a.cs7_instance = ss7->cfg.id;
 	msc->a.cs7_instance_valid = true;
 	enforce_standard_ssn(vty, &msc->a.msc_addr);
+	msc->a.msc_addr_name = talloc_strdup(msc, msc_addr_name);
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_msc_cs7_asp_proto,
+      cfg_msc_cs7_asp_proto_cmd,
+      "asp-protocol (m3ua|sua|ipa)",
+      "A interface protocol to use for this MSC)\n"
+      "MTP3 User Adaptation\n"
+      "SCCP User Adaptation\n"
+      "IPA Multiplex (SCCP Lite)\n")
+{
+	struct bsc_msc_data *msc = bsc_msc_data(vty);
+
+	msc->a.asp_proto = get_string_value(osmo_ss7_asp_protocol_vals, argv[0]);
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_net_msc_lcls_mode,
+      cfg_net_msc_lcls_mode_cmd,
+      "lcls-mode (disabled|mgw-loop|bts-loop)",
+      "Configure 3GPP LCLS (Local Call, Local Switch)\n"
+      "Disable LCLS for all calls of this MSC\n"
+      "Enable LCLS with looping traffic in MGW\n"
+      "Enable LCLS with looping traffic between BTS\n")
+{
+	struct bsc_msc_data *data = bsc_msc_data(vty);
+	data->lcls_mode = get_string_value(bsc_lcls_mode_names, argv[0]);
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_net_msc_lcls_mismtch,
+      cfg_net_msc_lcls_mismtch_cmd,
+      "lcls-codec-mismatch (allowed|forbidden)",
+      "Allow 3GPP LCLS (Local Call, Local Switch) when call legs use different codec/rate\n"
+      "Allow LCLS only only for calls that use the same codec/rate on both legs\n"
+      "Do not Allow LCLS for calls that use a different codec/rate on both legs\n")
+{
+	struct bsc_msc_data *data = bsc_msc_data(vty);
+
+	if (strcmp(argv[0], "allowed") == 0)
+		data->lcls_codec_mismatch_allow = true;
+	else
+		data->lcls_codec_mismatch_allow = false;
+
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_msc_mgw_x_osmo_ign,
+      cfg_msc_mgw_x_osmo_ign_cmd,
+      "mgw x-osmo-ign call-id",
+      MGCP_CLIENT_MGW_STR
+      "Set a (non-standard) X-Osmo-IGN header in all CRCX messages for RTP streams"
+      " associated with this MSC, useful for A/SCCPlite MSCs, since osmo-bsc cannot know"
+      " the MSC's chosen CallID. This is enabled by default for A/SCCPlite connections,"
+      " disabled by default for all others.\n"
+      "Send 'X-Osmo-IGN: C' to ignore CallID mismatches. See OsmoMGW.\n")
+{
+	struct bsc_msc_data *msc = bsc_msc_data(vty);
+	msc->x_osmo_ign |= MGCP_X_OSMO_IGN_CALLID;
+	msc->x_osmo_ign_configured = true;
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_msc_no_mgw_x_osmo_ign,
+      cfg_msc_no_mgw_x_osmo_ign_cmd,
+      "no mgw x-osmo-ign",
+      NO_STR
+      MGCP_CLIENT_MGW_STR
+      "Do not send X-Osmo-IGN MGCP header to this MSC\n")
+{
+	struct bsc_msc_data *msc = bsc_msc_data(vty);
+	msc->x_osmo_ign = 0;
+	msc->x_osmo_ign_configured = true;
+	return CMD_SUCCESS;
+}
+
+#define OSMUX_STR "RTP multiplexing\n"
+DEFUN(cfg_msc_osmux,
+      cfg_msc_osmux_cmd,
+      "osmux (on|off|only)",
+       OSMUX_STR "Enable OSMUX\n" "Disable OSMUX\n" "Only use OSMUX\n")
+{
+	struct bsc_msc_data *msc = bsc_msc_data(vty);
+	if (strcmp(argv[0], "off") == 0)
+		msc->use_osmux = OSMUX_USAGE_OFF;
+	else if (strcmp(argv[0], "on") == 0)
+		msc->use_osmux = OSMUX_USAGE_ON;
+	else if (strcmp(argv[0], "only") == 0)
+		msc->use_osmux = OSMUX_USAGE_ONLY;
+
 	return CMD_SUCCESS;
 }
 
 DEFUN(cfg_net_bsc_mid_call_text,
       cfg_net_bsc_mid_call_text_cmd,
       "mid-call-text .TEXT",
-      "Set the USSD notification to be send.\n" "Text to be sent\n")
+      "Set the USSD notification sent to running calls when switching from Grace to Off.\n"
+      "Text to be sent\n")
 {
 	struct osmo_bsc_data *data = osmo_bsc_data(vty);
 	char *txt = argv_concat(argv, argc, 0);
@@ -873,10 +831,12 @@ DEFUN(show_mscs,
 {
 	struct bsc_msc_data *msc;
 	llist_for_each_entry(msc, &bsc_gsmnet->bsc_data->mscs, entry) {
-		vty_out(vty, "MSC Nr: %d is connected: %d auth: %d.%s",
-			msc->nr,
-			msc->msc_con ? msc->msc_con->is_connected : -1,
-			msc->msc_con ? msc->msc_con->is_authenticated : -1,
+		vty_out(vty, "%d %s %s ",
+			msc->a.cs7_instance,
+			osmo_ss7_asp_protocol_name(msc->a.asp_proto),
+			osmo_sccp_inst_addr_name(msc->a.sccp, &msc->a.bsc_addr));
+		vty_out(vty, "%s%s",
+			osmo_sccp_inst_addr_name(msc->a.sccp, &msc->a.msc_addr),
 			VTY_NEWLINE);
 	}
 
@@ -954,25 +914,97 @@ DEFUN(logging_fltr_imsi,
 	struct log_target *tgt = osmo_log_vty2tgt(vty);
 	const char *imsi = argv[0];
 
-	bsc_subscr = bsc_subscr_find_by_imsi(bsc_gsmnet->bsc_subscribers, imsi);
+	if (!tgt)
+		return CMD_WARNING;
+
+	bsc_subscr = bsc_subscr_find_or_create_by_imsi(bsc_gsmnet->bsc_subscribers, imsi);
 
 	if (!bsc_subscr) {
-		vty_out(vty, "%%no subscriber with IMSI(%s)%s",
+		vty_out(vty, "%%failed to enable logging for subscriber with IMSI(%s)%s",
 			imsi, VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
 	log_set_filter_bsc_subscr(tgt, bsc_subscr);
+	/* log_set_filter has grabbed its own reference  */
+	bsc_subscr_put(bsc_subscr);
+
+	return CMD_SUCCESS;
+}
+
+static void dump_one_sub(struct vty *vty, struct bsc_subscr *bsub)
+{
+	vty_out(vty, " %15s  %08x  %5u  %d%s", bsub->imsi, bsub->tmsi, bsub->lac, bsub->use_count,
+		VTY_NEWLINE);
+}
+
+DEFUN(show_subscr_all,
+	show_subscr_all_cmd,
+	"show subscriber all",
+	SHOW_STR "Display information about subscribers\n" "All Subscribers\n")
+{
+	struct bsc_subscr *bsc_subscr;
+
+	vty_out(vty, " IMSI             TMSI      LAC    Use%s", VTY_NEWLINE);
+	/*           " 001010123456789  ffffffff  65534  1" */
+
+	llist_for_each_entry(bsc_subscr, bsc_gsmnet->bsc_subscribers, entry)
+		dump_one_sub(vty, bsc_subscr);
+
+	return CMD_SUCCESS;
+}
+
+#define LEGACY_STR "This command has no effect, it is kept to support legacy config files\n"
+
+DEFUN_DEPRECATED(cfg_net_msc_ping_time, cfg_net_msc_ping_time_cmd,
+      "timeout-ping ARG", LEGACY_STR "-\n")
+{
+	vty_out(vty, "%% timeout-ping / timeout-pong config is deprecated and has no effect%s",
+		VTY_NEWLINE);
+	return CMD_WARNING;
+}
+
+ALIAS_DEPRECATED(cfg_net_msc_ping_time, cfg_net_msc_no_ping_time_cmd,
+      "no timeout-ping [ARG]", NO_STR LEGACY_STR "-\n");
+
+ALIAS_DEPRECATED(cfg_net_msc_ping_time, cfg_net_msc_pong_time_cmd,
+      "timeout-pong ARG", LEGACY_STR "-\n");
+
+DEFUN_DEPRECATED(cfg_net_msc_dest, cfg_net_msc_dest_cmd,
+      "dest A.B.C.D <1-65000> <0-255>", LEGACY_STR "-\n" "-\n" "-\n")
+{
+	vty_out(vty, "%% dest config is deprecated and has no effect%s", VTY_NEWLINE);
+	return CMD_WARNING;
+}
+
+ALIAS_DEPRECATED(cfg_net_msc_dest, cfg_net_msc_no_dest_cmd,
+      "no dest A.B.C.D <1-65000> <0-255>", NO_STR LEGACY_STR "-\n" "-\n" "-\n");
+
+DEFUN(cfg_net_msc_amr_octet_align,
+      cfg_net_msc_amr_octet_align_cmd,
+      "amr-payload (octet-aligned|bandwith-efficient",
+      "Set AMR payload framing mode\n"
+      "payload fields aligned on octet boundaries\n"
+      "payload fields packed (AoIP)\n")
+{
+	struct bsc_msc_data *data = bsc_msc_data(vty);
+
+	if (strcmp(argv[0], "octet-aligned") == 0)
+		data->amr_octet_aligned = true;
+	else if (strcmp(argv[0], "bandwith-efficient") == 0)
+		data->amr_octet_aligned = false;
+
 	return CMD_SUCCESS;
 }
 
 int bsc_vty_init_extra(void)
 {
+	struct gsm_network *net = bsc_gsmnet;
+
 	install_element(CONFIG_NODE, &cfg_net_msc_cmd);
 	install_element(CONFIG_NODE, &cfg_net_bsc_cmd);
 
 	install_node(&bsc_node, config_write_bsc);
-	vty_install_default(BSC_NODE);
 	install_element(BSC_NODE, &cfg_net_bsc_mid_call_text_cmd);
 	install_element(BSC_NODE, &cfg_net_bsc_mid_call_timeout_cmd);
 	install_element(BSC_NODE, &cfg_net_rf_socket_cmd);
@@ -984,10 +1016,6 @@ int bsc_vty_init_extra(void)
 	install_element(BSC_NODE, &cfg_bsc_no_acc_lst_name_cmd);
 
 	install_node(&msc_node, config_write_msc);
-	vty_install_default(MSC_NODE);
-	install_element(MSC_NODE, &cfg_net_bsc_token_cmd);
-	install_element(MSC_NODE, &cfg_net_bsc_key_cmd);
-	install_element(MSC_NODE, &cfg_net_bsc_no_key_cmd);
 	install_element(MSC_NODE, &cfg_net_bsc_ncc_cmd);
 	install_element(MSC_NODE, &cfg_net_bsc_mcc_cmd);
 	install_element(MSC_NODE, &cfg_net_bsc_lac_cmd);
@@ -996,11 +1024,6 @@ int bsc_vty_init_extra(void)
 	install_element(MSC_NODE, &cfg_net_bsc_codec_list_cmd);
 	install_element(MSC_NODE, &cfg_net_msc_dest_cmd);
 	install_element(MSC_NODE, &cfg_net_msc_no_dest_cmd);
-	install_element(MSC_NODE, &cfg_net_msc_no_ping_time_cmd);
-	install_element(MSC_NODE, &cfg_net_msc_ping_time_cmd);
-	install_element(MSC_NODE, &cfg_net_msc_pong_time_cmd);
-	install_element(MSC_NODE, &cfg_net_msc_advanced_ping_cmd);
-	install_element(MSC_NODE, &cfg_no_net_msc_advanced_ping_cmd);
 	install_element(MSC_NODE, &cfg_net_msc_welcome_ussd_cmd);
 	install_element(MSC_NODE, &cfg_net_msc_no_welcome_ussd_cmd);
 	install_element(MSC_NODE, &cfg_net_msc_lost_ussd_cmd);
@@ -1018,19 +1041,34 @@ int bsc_vty_init_extra(void)
 	install_element(MSC_NODE, &cfg_net_msc_amr_5_90_cmd);
 	install_element(MSC_NODE, &cfg_net_msc_amr_5_15_cmd);
 	install_element(MSC_NODE, &cfg_net_msc_amr_4_75_cmd);
+	install_element(MSC_NODE, &cfg_net_msc_amr_octet_align_cmd);
+	install_element(MSC_NODE, &cfg_net_msc_lcls_mode_cmd);
+	install_element(MSC_NODE, &cfg_net_msc_lcls_mismtch_cmd);
 	install_element(MSC_NODE, &cfg_msc_acc_lst_name_cmd);
 	install_element(MSC_NODE, &cfg_msc_no_acc_lst_name_cmd);
 	install_element(MSC_NODE, &cfg_msc_cs7_bsc_addr_cmd);
 	install_element(MSC_NODE, &cfg_msc_cs7_msc_addr_cmd);
+	install_element(MSC_NODE, &cfg_msc_cs7_asp_proto_cmd);
+
+	/* Deprecated: ping time config, kept to support legacy config files. */
+	install_element(MSC_NODE, &cfg_net_msc_no_ping_time_cmd);
+	install_element(MSC_NODE, &cfg_net_msc_ping_time_cmd);
+	install_element(MSC_NODE, &cfg_net_msc_pong_time_cmd);
 
 	install_element_ve(&show_statistics_cmd);
 	install_element_ve(&show_mscs_cmd);
 	install_element_ve(&show_pos_cmd);
 	install_element_ve(&logging_fltr_imsi_cmd);
+	install_element_ve(&show_subscr_all_cmd);
 
 	install_element(ENABLE_NODE, &gen_position_trap_cmd);
 
 	install_element(CFG_LOG_NODE, &logging_fltr_imsi_cmd);
 
+	mgcp_client_vty_init(net, MSC_NODE, net->mgw.conf);
+	install_element(MSC_NODE, &cfg_msc_mgw_x_osmo_ign_cmd);
+	install_element(MSC_NODE, &cfg_msc_no_mgw_x_osmo_ign_cmd);
+	install_element(MSC_NODE, &cfg_msc_osmux_cmd);
+
 	return 0;
 }
